/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s): 
 *   Justin Bradford <jab@atdot.org> (original author of nsDigestAuth.cpp)
 *   An-Cheng Huang <pach@cs.cmu.edu>
 */

#include <stdlib.h>
#include "nsHttp.h"
#include "nsHttpDigestAuth.h"
#include "plbase64.h"
#include "plstr.h"
#include "prmem.h"
#include "nsIServiceManager.h"
#include "nsString.h"
#include "nsIHttpChannel.h"

//-----------------------------------------------------------------------------
// nsHttpDigestAuth <public>
//-----------------------------------------------------------------------------

nsHttpDigestAuth::nsHttpDigestAuth()
{
  NS_INIT_ISUPPORTS();

  nsresult rv;
  mVerifier = do_GetService(SIGNATURE_VERIFIER_CONTRACTID, &rv);
  if (NS_FAILED(rv)) {
    LOG(("nsHttpDigestAuth: No signature_verifier available\n"));
    gotVerifier = PR_FALSE;
  } else {
    LOG(("nsHttpDigestAuth: Got signature_verifier\n"));
    gotVerifier = PR_TRUE;
  }
}

nsHttpDigestAuth::~nsHttpDigestAuth()
{
}

//-----------------------------------------------------------------------------
// nsHttpDigestAuth::nsISupports
//-----------------------------------------------------------------------------

NS_IMPL_ISUPPORTS1(nsHttpDigestAuth, nsIHttpAuthenticator);

//-----------------------------------------------------------------------------
// nsHttpDigestAuth <protected>
//-----------------------------------------------------------------------------

nsresult
nsHttpDigestAuth::MD5Hash(char *buf, PRUint32 len)
{
  if (!gotVerifier) {
    return NS_ERROR_UNEXPECTED; // maybe not
  }

  nsresult rv;
  HASHContextStr *hid;
  unsigned char *chash = (unsigned char *) PR_MALLOC(DIGEST_LENGTH);
  PRUint32 clen;
  rv = mVerifier->HashBegin(nsISignatureVerifier::MD5, &hid);
  if (NS_FAILED(rv))
    goto hasherror;
  rv = mVerifier->HashUpdate(hid, buf, len);
  if (NS_FAILED(rv))
    goto hasherror;
  rv = mVerifier->HashEnd(hid, &chash, &clen, DIGEST_LENGTH);
  if (NS_FAILED(rv))
    goto hasherror;
  nsCRT::memcpy(mHashBuf, chash, DIGEST_LENGTH);
  PR_FREEIF(chash);
  return NS_OK;

hasherror:
  PR_FREEIF(chash);
  return NS_ERROR_UNEXPECTED;
}

//-----------------------------------------------------------------------------
// nsHttpDigestAuth::nsIHttpAuthenticator
//-----------------------------------------------------------------------------

nsresult
nsHttpDigestAuth::NewCredentialRequired(nsIHttpChannel *httpChannel,
                                        const char *creds,
                                        PRBool *retval)
{
    (*retval) = PR_TRUE;
    return NS_OK;
}

nsresult
nsHttpDigestAuth::GenerateCredentials(nsIHttpChannel *httpChannel,
                                     const char *challenge,
                                     const PRUnichar *username,
                                     const PRUnichar *password,
                                     char **creds)

{
  LOG(("nsHttpDigestAuth::GenerateCredentials [challenge=%s]\n", challenge));
  printf("nsHttpDigestAuth::GenerateCredentials [challenge=%s]\n", challenge);

  PRBool isDigestAuth = !PL_strncasecmp(challenge, "digest ", 7);
  NS_ENSURE_TRUE(isDigestAuth, NS_ERROR_UNEXPECTED);

  NS_ENSURE_ARG_POINTER(creds);

  nsCAutoString cUser, cPass;
  cUser.AssignWithConversion(username);
  if (password) {
    cPass.AssignWithConversion(password);
  }

  nsresult rv;
  nsIURI *iURI;
  char *http_uri;
  char *http_method;
  rv = httpChannel->GetURI(&iURI);
  if (NS_SUCCEEDED(rv)) {
    iURI->GetPath(&http_uri);
  }
  rv = httpChannel->GetRequestMethod(&http_method);
  if (!NS_SUCCEEDED(rv)) {
    LOG(("nsHttpDigestAuth::GenerateCredentials [GetRequestMethod failed]\n"));
  }

  nsCAutoString realm, domain, nonce, opaque;
  PRBool stale;
  PRUint16 algorithm, qop;

  rv = ParseChallenge(challenge, realm, domain, nonce, opaque,
                      &stale, &algorithm, &qop);
  if (NS_FAILED(rv)) {
    LOG(("nsHttpDigestAuth::GenerateCredentials [ParseChallenge failed]\n"));
    return NS_ERROR_INVALID_ARG;
  }

  char ha1_digest[EXPANDED_DIGEST_LENGTH+1];
  char ha2_digest[EXPANDED_DIGEST_LENGTH+1];
  char response_digest[EXPANDED_DIGEST_LENGTH+1];
  char upload_data_digest[EXPANDED_DIGEST_LENGTH+1];

  if (qop & QOP_AUTH_INT) {
    // we do not support auth-int "quality of protection" currently
    qop &= ~QOP_AUTH_INT;

#ifdef DEBUGGING
    printf("no support for Digest authentication with data integrity quality of protection\n");
#endif

    /* TODO: to support auth-int, we need to get an MD5 digest of
     * TODO: the data uploaded with this request.
     * TODO: however, i am not sure how to read in the file in without
     * TODO: disturbing the channel''s use of it. do i need to copy it 
     * TODO: somehow?
     */
#if 0
    if (http_channel != nsnull)
    {
      nsIInputStream * upload;
      http_channel->GetUploadStream(&upload);
      if (upload) {
        char * upload_buffer;
        int upload_buffer_length = 0;
        //TODO: read input stream into buffer
        const char * digest = (const char*)
        nsNetwerkMD5Digest(upload_buffer, upload_buffer_length);
        ExpandToHex(digest, upload_data_digest);
        NS_RELEASE(upload);
      }
    }
#endif
  }

  if (!(algorithm & ALGO_MD5 || algorithm & ALGO_MD5_SESS)) {
    // they asked only for algorithms that we do not support
#ifdef DEBUGGING
    printf("unsupported algorithm requested by Digest authentication\n");
#endif
    return NS_ERROR_INVALID_ARG;
  }

  /* the following are for increasing security (but we currently do not use)
   * see RFC 2617 for more information
   *
   * this allows the server to keep track auth challenges
   * (to help prevent spoofing). ideally, we would increase this
   * count every time we see a specific nonce.
   */
  char nonce_count[NONCE_COUNT_LENGTH+1] = "00000001"; // in hex
  /* this lets the client verify the server response (via a server
   * returned Authentication-Info header). also used for session info.
   */
  // this is random, and should be changed
  nsCAutoString cnonce("0123456789abcdef"); 

  rv = CalculateHA1(cUser, cPass, realm, algorithm, nonce, cnonce, 
                    ha1_digest);
  if (NS_FAILED(rv))
    return NS_ERROR_UNEXPECTED;
  rv = CalculateHA2(http_method, http_uri, qop, upload_data_digest, 
                    ha2_digest);
  if (NS_FAILED(rv))
    return NS_ERROR_UNEXPECTED;
  rv = CalculateResponse(ha1_digest, ha2_digest, nonce, qop, nonce_count,
                         cnonce, response_digest);
  if (NS_FAILED(rv))
    return NS_ERROR_UNEXPECTED;

  nsCAutoString authString("Digest ");

  authString += "username=\"";
  authString += cUser;
  authString += "\", realm=\"";
  authString += realm;
  authString += "\", nonce=\"";
  authString += nonce;
  authString += "\", uri=\"";
  authString += http_uri;
  if (algorithm & ALGO_SPECIFIED) {
    authString += "\", algorithm=";
    if (algorithm & ALGO_MD5_SESS)
      authString += "MD5-sess";
    else
      authString += "MD5";
  } else {
    authString += "\"";
  }
  authString += ", response=\"";
  authString += response_digest;

  if (!opaque.IsEmpty()) {
    authString += "\", opaque=\"";
    authString += opaque;
  }

  if (qop) {
    authString += "\", qop=";
    if (qop & QOP_AUTH_INT)
      authString += "auth-int";
    else
      authString += "auth";
    authString += ", nc=";
    authString += nonce_count;
    authString += ", cnonce=\"";
    authString += cnonce;
  }
  authString += "\"";

  CRTFREEIF(http_uri);
  CRTFREEIF(http_method);

  *creds = authString.ToNewCString();
  LOG(("nsHttpDigestAuth::GenerateCredentials [authString=%s]\n",
        *creds));
  return NS_OK;
}

nsresult
nsHttpDigestAuth::CalculateResponse(const char * ha1_digest,
                                const char * ha2_digest,
                                nsCAutoString & nonce,
                                PRUint16 qop,
                                const char * nonce_count,
                                nsCAutoString & cnonce,
                                char * result)
{
  PRInt16 len = EXPANDED_DIGEST_LENGTH + EXPANDED_DIGEST_LENGTH
    + nonce.Length() + 2;

  if (qop & QOP_AUTH || qop & QOP_AUTH_INT) {
    len += cnonce.Length() + NONCE_COUNT_LENGTH + 3;
    if (qop & QOP_AUTH_INT)
      len += 8; // length of "auth-int"
    else
      len += 4; // length of "auth"
  }

  char * contents = (char*)nsMemory::Alloc(len);
  if (contents == nsnull) 
    return NS_ERROR_OUT_OF_MEMORY;

  nsCRT::memcpy(contents, ha1_digest, EXPANDED_DIGEST_LENGTH);
  len = EXPANDED_DIGEST_LENGTH;
  contents[len++] = ':';
  nsCRT::memcpy(contents+len, nonce, nonce.Length());
  len += nonce.Length();
  contents[len++] = ':';

  if (qop & QOP_AUTH || qop & QOP_AUTH_INT) {
    // bug
    //nsCRT::memcpy(contents, nonce_count, NONCE_COUNT_LENGTH);
    nsCRT::memcpy(contents+len, nonce_count, NONCE_COUNT_LENGTH);
    len += NONCE_COUNT_LENGTH;
    contents[len++] = ':';
    nsCRT::memcpy(contents+len, cnonce, cnonce.Length());
    len += cnonce.Length();
    contents[len++] = ':';
    if (qop & QOP_AUTH_INT) {
      nsCRT::memcpy(contents+len, "auth-int:", 9);
      len += 9;
    } else {
      nsCRT::memcpy(contents+len, "auth:", 5);
      len += 5;
    }

  }

  nsCRT::memcpy(contents+len, ha2_digest, EXPANDED_DIGEST_LENGTH);
  len += EXPANDED_DIGEST_LENGTH;

  nsresult rv;
  rv = MD5Hash(contents, len);
  if (NS_FAILED(rv))
    return rv;

  nsMemory::Free(contents);

  return ExpandToHex(mHashBuf, result);
}

nsresult
nsHttpDigestAuth::ExpandToHex(const char * digest, char * result)
{
  PRInt16 index, value;

  for (index = 0; index < DIGEST_LENGTH; index++) {

    value = (digest[index] >> 4) & 0xf;
    if (value < 10)
      result[index*2] = value + '0';
    else
      result[index*2] = value - 10 + 'a';

    value = digest[index] & 0xf;
    if (value < 10)
      result[(index*2)+1] = value + '0';
    else
      result[(index*2)+1] = value - 10 + 'a';

  }

  result[EXPANDED_DIGEST_LENGTH] = 0;

  return NS_OK;
}

nsresult
nsHttpDigestAuth::CalculateHA1(nsCAutoString & username,
                           nsCAutoString & password,
                           nsCAutoString & realm,
                           PRUint16 algorithm,
                           nsCAutoString & nonce,
                           nsCAutoString & cnonce,
                           char * result)
{
  PRInt16 len = username.Length() + password.Length() + realm.Length() + 2;
  if (algorithm & ALGO_MD5_SESS) {
    PRInt16 exlen = EXPANDED_DIGEST_LENGTH +
                    nonce.Length() + cnonce.Length() + 2;
    if (exlen > len) len = exlen;
  }
  char * contents = (char*)nsMemory::Alloc(len);
  if (contents == nsnull)
    return NS_ERROR_OUT_OF_MEMORY;

  nsCRT::memcpy(contents, username, username.Length());
  len = username.Length();
  contents[len++] = ':';
  nsCRT::memcpy(contents+len, realm, realm.Length());
  len += realm.Length();
  contents[len++] = ':';
  nsCRT::memcpy(contents+len, password, password.Length());
  len += password.Length();

  nsresult rv;
  rv = MD5Hash(contents, len);
  if (NS_FAILED(rv))
    return rv;

  if (algorithm & ALGO_MD5_SESS) {
    char part1[EXPANDED_DIGEST_LENGTH+1];
    ExpandToHex(mHashBuf, part1);
    nsCRT::memcpy(contents, part1, EXPANDED_DIGEST_LENGTH);
    len = EXPANDED_DIGEST_LENGTH;
    contents[len++] = ':';
    nsCRT::memcpy(contents+len, nonce, nonce.Length());
    len += nonce.Length();
    contents[len++] = ':';
    nsCRT::memcpy(contents+len, cnonce, cnonce.Length());
    len += cnonce.Length();

    rv = MD5Hash(contents, len);
    if (NS_FAILED(rv))
      return rv;
  }

  nsMemory::Free(contents);

  return ExpandToHex(mHashBuf, result);
}

nsresult
nsHttpDigestAuth::CalculateHA2(const char * http_method,
                           const char * http_uri,
                           PRUint16 qop,
                           const char * body_digest,
                           char * result)
{
  PRInt16 http_method_len = nsCRT::strlen(http_method);
  PRInt16 http_uri_len = nsCRT::strlen(http_uri);
  PRInt16 len = http_method_len + http_uri_len + 1;

  if (qop & QOP_AUTH_INT) {
    len += EXPANDED_DIGEST_LENGTH + 1;
  }

  char * contents = (char*)nsMemory::Alloc(len);
  if (contents == nsnull)
    return NS_ERROR_OUT_OF_MEMORY;

  nsCRT::memcpy(contents, http_method, http_method_len);
  len = http_method_len;
  contents[len++] = ':';
  nsCRT::memcpy(contents+len, http_uri, http_uri_len);
  len += http_uri_len;

  if (qop & QOP_AUTH_INT) {

    contents[len++] = ':';
    nsCRT::memcpy(contents+len, body_digest, EXPANDED_DIGEST_LENGTH);
    len += EXPANDED_DIGEST_LENGTH;

  }

  nsresult rv;
  rv = MD5Hash(contents, len);
  if (NS_FAILED(rv))
    return rv;

  nsMemory::Free(contents);

  return ExpandToHex(mHashBuf, result);
}

nsresult
nsHttpDigestAuth::ParseChallenge(const char * challenge,
                             nsCAutoString & realm,
                             nsCAutoString & domain,
                             nsCAutoString & nonce,
                             nsCAutoString & opaque,
                             PRBool * stale,
                             PRUint16 * algorithm,
                             PRUint16 * qop)
{
  PRInt16 pos = 7; //first 7 characters are "Digest "

  realm = "";
  domain = "";
  nonce = "";
  opaque = "";
  *stale = PR_FALSE;
  *algorithm = ALGO_MD5; // default is MD5
  *qop = 0;

  for (;;) {
    while (challenge[pos] && (challenge[pos] == ',' ||
          nsCRT::IsAsciiSpace(challenge[pos]))) 
      pos++;
    if (!challenge[pos]) 
      break;

    // name
    PRInt16 nameStart = pos;
    while (challenge[pos] && !nsCRT::IsAsciiSpace(challenge[pos]) &&
        challenge[pos] != '=') 
      pos++;
    if (!challenge[pos]) 
      return NS_ERROR_INVALID_ARG;
    PRInt16 nameLength = pos - nameStart;

    while (challenge[pos] && (nsCRT::IsAsciiSpace(challenge[pos]) ||
          challenge[pos] == '=')) 
      pos++;
    if (!challenge[pos]) 
      return NS_ERROR_INVALID_ARG;

    PRBool quoted = PR_FALSE;
    if (challenge[pos] == '"') {
      pos++;
      quoted = PR_TRUE;
    }

    // value
    PRInt16 valueStart = pos;
    PRInt16 valueLength = 0;
    if (quoted) {
      while (challenge[pos] && challenge[pos] != '"') 
        pos++;
      if (challenge[pos] != '"') 
        return NS_ERROR_INVALID_ARG;
      valueLength = pos - valueStart;
      pos++;
    } else {
      while (challenge[pos] && !nsCRT::IsAsciiSpace(challenge[pos])
          && challenge[pos] != ',') 
        pos++;
      // bug??
      //if (!challenge[pos]) return NS_ERROR_INVALID_ARG;
      valueLength = pos - valueStart;
    }

    // extract information
    if (nameLength == 5 &&
        nsCRT::strncasecmp(challenge+nameStart, "realm", 5) == 0)
    {
      realm.Assign(challenge+valueStart, valueLength);
    }
    else if (nameLength == 6 &&
        nsCRT::strncasecmp(challenge+nameStart, "domain", 6) == 0)
    {
      domain.Assign(challenge+valueStart, valueLength);
    }
    else if (nameLength == 5 &&
        nsCRT::strncasecmp(challenge+nameStart, "nonce", 5) == 0)
    {
      nonce.Assign(challenge+valueStart, valueLength);
    }
    else if (nameLength == 6 &&
        nsCRT::strncasecmp(challenge+nameStart, "opaque", 6) == 0)
    {
      opaque.Assign(challenge+valueStart, valueLength);
    }
    else if (nameLength == 5 &&
        nsCRT::strncasecmp(challenge+nameStart, "stale", 5) == 0)
    {
      if (nsCRT::strncasecmp(challenge+valueStart, "true", 4) == 0)
        *stale = PR_TRUE;
      else
        *stale = PR_FALSE;
    }
    else if (nameLength == 9 &&
        nsCRT::strncasecmp(challenge+nameStart, "algorithm", 9) == 0)
    {
      // we want to clear the default, so we use = not |= here
      *algorithm = ALGO_SPECIFIED;
      if (valueLength == 3 &&
          nsCRT::strncasecmp(challenge+valueStart, "MD5", 3) == 0)
        *algorithm |= ALGO_MD5;
      else if (valueLength == 8 &&
          nsCRT::strncasecmp(challenge+valueStart, "MD5-sess", 8) == 0)
        *algorithm |= ALGO_MD5_SESS;
    }
    else if (nameLength == 3 &&
        nsCRT::strncasecmp(challenge+nameStart, "qop", 3) == 0)
    {
      PRInt16 ipos = valueStart;
      while (ipos < valueStart+valueLength) {
        while (ipos < valueStart+valueLength &&
               (nsCRT::IsAsciiSpace(challenge[ipos]) ||
                challenge[ipos] == ',')) 
          ipos++;
        PRInt16 algostart = ipos;
        while (ipos < valueStart+valueLength &&
               !nsCRT::IsAsciiSpace(challenge[ipos]) &&
               challenge[ipos] != ',') 
          ipos++;
        // bug??
        //if (ipos >= valueStart+valueLength) return NS_ERROR_INVALID_ARG;
        if ((ipos - algostart) == 4 &&
            nsCRT::strncasecmp(challenge+algostart, "auth", 4) == 0)
          *qop |= QOP_AUTH;
        else if ((ipos - algostart) == 8 &&
            nsCRT::strncasecmp(challenge+algostart, "auth-int", 8) == 0)
          *qop |= QOP_AUTH_INT;
      }
    }
  }
  return NS_OK;
}
